
from abnamro import services
from abnamro.schema import *


class Action(Schema):
	name: StringField("name")
	user_action_indicator: StringField("userActionIndicator")


class Mutation(Schema):
	amount: FloatField("amount")
	currency_iso_code: StringField("currencyIsoCode")
	balance_after_mutation: FloatField("balanceAfterMutation")
	debit_credit: StringField("debitCredit", choices=["DEBIT", "CREDIT"])
	indicator_digital_invoice: BooleanField("indicatorDigitalInvoice")
	payment_status: StringField("paymentStatus", choices=["DEFAULT", "INITIAL", "APPROVED", "PENDING", "REJECTED"])
	status_timestamp: TimestampField("statusTimestamp")

	mutation_code: StringField("mutationCode")
	description_lines: StringField("descriptionLines", repeated=True)
	source_inquiry_number: StringField("sourceInquiryNumber")
	transaction_timestamp: StringField("transactionTimestamp")

	account_type: StringField("accountNumberType")
	account_number: StringField("accountNumber")
	counter_account_name: StringField("counterAccountName")
	counter_account_type: StringField("counterAccountType")
	counter_account_number: StringField("counterAccountNumber")

	transaction_date: TimestampField("transactionDate")
	value_date: TimestampField("valueDate")
	book_date: TimestampField("bookDate")


class MutationObject(Schema):
	actions: ObjectField("actions", Action, repeated=True)
	mutation: ObjectField("mutation", Mutation)


class MutationList(Schema):
	clear_cache_indicator: BooleanField("clearCacheIndicator")
	page_token: StringField("lastMutationKey")
	mutations: ObjectField("mutations", MutationObject, repeated=True)


class MutationsManager:
	def __init__(self, client):
		self.service = services.MutationsService(client)
	
	async def search(
		self, iban, *, page_size=20, page_token=None, text=None, amount_from=None,
		amount_to=None, book_date_from=None, book_date_to=None
	):
		cd_indicator_amount_from = None
		cd_indicator_amount_to = None

		if amount_from is not None:
			cd_indicator_amount_from = "CREDIT" if amount_from > 0 else "DEBIT"
			amount_from = abs(amount_from)
		
		if amount_to is not None:
			cd_indicator_amount_to = "CREDIT" if amount_to > 0 else "DEBIT"
			amount_to = abs(amount_to)

		response = await self.service.get_mutations(
			iban, page_size, "BASIC", page_token, None,
			text, amount_from, amount_to, cd_indicator_amount_from,
			cd_indicator_amount_to, book_date_from, book_date_to
		)
		return MutationList.create(response["mutationsList"])




from abnamro import services
from abnamro.schema import *


class Balance(Schema):
	amount: FloatField("amount")
	currency_code: StringField("currencyCode")


class Product(Schema):
	resouce_type: StringField("resourceType")
	id: IntegerField("id")
	building_block_id: IntegerField("buildingBlockId")
	name: StringField("name")
	product_group: StringField("productGroup")
	account_type: StringField("accountType")
	transfer_options: StringField("transferOptions", repeated=True)
	credit_account: BooleanField("creditAccount")


class Customer(Schema):
	appearance_type: StringField("appearanceType")
	bc_number: IntegerField("bcNumber")
	interpay_name: StringField("interpayName")


class Action(Schema):
	name: StringField("name")


class ParentContract(Schema):
	id: StringField("id")


class Contract(Schema):
	resource_type: StringField("resourceType")
	id: StringField("id")
	account_number: StringField("accountNumber")
	contract_number: StringField("contractNumber")
	chid: StringField("chid")
	concerning: StringField("concerning")
	is_blocked: BooleanField("isBlocked")
	status: StringField("status")
	balance: ObjectField("balance", Balance)
	product: ObjectField("product", Product)
	customer: ObjectField("customer", Customer)
	actions: ObjectField("actions", Action, repeated=True)
	sequence_number: IntegerField("sequenceNumber")
	parent_contract: ObjectField("parentContract", ParentContract)


class AccountsManager:
	def __init__(self, client):
		self.service = services.AccountsService(client)
	
	async def list(self, *, product_groups=None, product_building_blocks=None, include_actions=None,
		include_action_names=None, exclude_blocked=None, exclude_status=None,
		bc_number=None, contract_ids=None
	):
		response = await self.service.get_contracts(
			product_groups, product_building_blocks, include_actions,
			include_action_names, exclude_blocked, exclude_status,
			bc_number, contract_ids
		)
		return [Contract.create(contract["contract"]) for contract in response["contractList"]]




APP_ID_ANDROID = "ANDROID_APP"
APP_ID_IPHONE = "IPHONE_APP"
APP_ID_SIMPLE_BANKING = "SIMPLE_BANKING"

ACCESS_EDENTIFIER1 = "EDENTIFIER1"
ACCESS_EDENTIFIER2_CONNECTED = "EDENTIFIER2_CONNECTED"
ACCESS_EDENTIFIER2_UNCONNECTED = "EDENTIFIER2_UNCONNECTED"
ACCESS_SOFTTOKEN = "SOFTTOKEN"
ACCESS_BOUNDDEVICE_USERPIN = "BOUNDDEVICE_USERPIN"
ACCESS_BOUNDDEVICE_TOUCHIDPIN = "BOUNDDEVICE_TOUCHIDPIN"
ACCESS_SESSIONHANDOVER = "SESSIONHANDOVER"
ACCESS_OOBGENERIC = "OOBGENERIC"




from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA

from abnamro import streams


def parse_fields(data):
	fields = {}
	stream = streams.StreamIn(data, ">")
	while not stream.eof():
		type = stream.u8()
		size = stream.u16()
		fields[type] = stream.read(size)
	return fields

def serialize_fields(fields):
	stream = streams.StreamOut(">")
	for key, value in fields.items():
		stream.u8(key)
		stream.u16(len(value))
		stream.write(value)
	return stream.get()

def solve_login_challenge(challenge, username, password):
	fields = parse_fields(challenge)
	data = serialize_fields({
		1: b"1",
		2: fields[2],
		3: fields[3],
		8: username.encode(),
		9: password.encode(),
		0: b""
	})
	key = RSA.construct((int.from_bytes(fields[4], "big"), int.from_bytes(fields[5], "big")))
	cipher = PKCS1_v1_5.new(key)
	return cipher.encrypt(data).hex()




from abnamro import services
from abnamro.schema import *


class DebitCardId(Schema):
	account_number: StringField("accountNumber")
	card_number: StringField("cardNumber")
	iban: StringField("ibanNumber")


class Duration(Schema):
	start_date: DateTimeField("startDate", "%Y-%m-%d")
	end_date: DateTimeField("endDate", "%Y-%m-%d")
	type: StringField("type", choices=["PERMANENT", "TEMPORARY"])


class LimitObject(Schema):
	amount: IntegerField("amount")
	currency: StringField("currencyCode")


class Limit(Schema):
	duration: ObjectField("duration", Duration)
	limit: ObjectField("limit", LimitObject)


class GeoProfile(Schema):
	type: StringField("geoProfile", choices=["EUROPE", "WORLD"])
	duration: ObjectField("duration", Duration)


class ValidationData(Schema):
	maximum_atm_limit: ObjectField("maximumATMLimit", LimitObject)
	maximum_pos_limit: ObjectField("maximumPOSLimit", LimitObject)
	minimum_atm_limit: ObjectField("minimumATMLimit", LimitObject)
	minimum_pos_limit: ObjectField("minimumPOSLimit", LimitObject)
	valid_geo_profile_types: StringField("validGeoProfileTypes", choices=["EUROPE", "WORLD"], repeated=True)


class DebitCard(Schema):
	id: ObjectField("id", DebitCardId)
	status: StringField("cardStatus", choices=["ACTIVE", "INACTIVE", "BLOCKED", "EXPIRED"])
	product_code: StringField("productCode")
	product_name: StringField("productName")
	owner_name: StringField("cardOwnerName")
	atm_limit: ObjectField("atmLimit", Limit)
	pos_limit: ObjectField("posLimit", Limit)
	geo_profile: ObjectField("geoProfile", GeoProfile)
	future_atm_limit: ObjectField("futureAtmLimit", Limit)
	future_pos_limit: ObjectField("futurePosLimit", Limit)
	future_geo_profile: ObjectField("futureGeoProfile", GeoProfile)
	is_atm_limit_change_allowed: BooleanField("isATMLimitChangeAllowed")
	is_pos_limit_change_allowed: BooleanField("isPOSLimitChangeAllowed")
	is_geo_profile_change_allowed: BooleanField("isGeoProfileChangeAllowed")
	is_atm_limit_update_in_progress: BooleanField("isATMLimitUpdateInProgress")
	is_pos_limit_update_in_progress: BooleanField("isPOSLimitUpdateInProgress")
	is_geo_profile_update_in_progress: BooleanField("isGeoProfileUpdateInProgress")
	is_manage_limit_authorized: BooleanField("isManageLimitAuthorised")
	is_manage_geo_profile_authorized: BooleanField("isManageGeoProfileAuthorised")
	last_mutation_date: TimestampField("lastMutationDate")
	last_reissue_date: TimestampField("lastReissueDate")
	last_block_date: TimestampField("lastBlockDate")
	validation_data: ObjectField("validationData", ValidationData)


class DebitCardsManager:
	def __init__(self, client):
		self.service = services.DebitCardsService(client)
	
	async def list(self):
		response = await self.service.get_debit_cards()
		cards = response["debitcardList"]["debitcards"]
		return [DebitCard.create(card["debitcard"]) for card in cards]
	
	async def get(self, account_number, card_number):
		card = await self.service.get_debit_card("%i;%i" %(account_number, card_number))
		return DebitCard.create(card["debitcard"])




import datetime
import inspect


class ValidationError(Exception):
	pass


class Field:
	def __init__(self, name, *, required=False, repeated=False):
		self.name = name
		self.required = required
		self.repeated = repeated
	
	def encode(self, value): return value
	def parse(self, value): return value


class BooleanField(Field):
	def parse(self, value):
		if not isinstance(value, bool):
			raise ValidationError("Expected a boolean, got %s" %type(value))
		return value


class IntegerField(Field):
	def parse(self, value):
		if not isinstance(value, int):
			raise ValidationError("Expected an integer, got %s" %type(value))
		return value


class FloatField(Field):
	def parse(self, value):
		if not isinstance(value, float):
			raise ValidationError("Expected a float, got %s" %type(value))
		return value


class StringField(Field):
	def __init__(self, name, *, choices=None, **kwargs):
		super().__init__(name, **kwargs)
		self.choices = choices
	
	def parse(self, value):
		if not isinstance(value, str):
			raise ValidationError("Expected a string, got %s" %type(value))
		if self.choices is not None and value not in self.choices:
			raise ValidationError("Received unexpected enum value '%s'" %value)
		return value


class DateTimeField(Field):
	def __init__(self, name, format, **kwargs):
		super().__init__(name, **kwargs)
		self.format = format
	
	def encode(self, value):
		return value.strftime(self.format)
	
	def parse(self, value):
		if not isinstance(value, str):
			raise ValidationError("Expected a string, got %s" %type(value))
		return datetime.datetime.strptime(value, self.format)


class TimestampField(Field):
	def encode(self, value):
		try:
			return int(value.timestamp() * 1000)
		except ValueError:
			return -62135773200000
	
	def parse(self, value):
		if not isinstance(value, int):
			raise ValidationError("Expected an integer, got %s" %type(value))
		
		try:
			return datetime.datetime.fromtimestamp(value / 1000)
		except ValueError:
			return datetime.datetime.min


class ObjectField(Field):
	def __init__(self, name, schema, **kwargs):
		super().__init__(name, **kwargs)
		self.schema = schema
	
	def encode(self, value):
		return value.encode()
	
	def parse(self, value):
		if not isinstance(value, dict):
			raise ValidationError("Expected a dict, got %s" %type(value))
		return self.schema.create(value)


class Schema:
	def __init__(self):
		self.reset()
	
	def __repr__(self):
		fields = []
		for name, field in inspect.get_annotations(self.__class__).items():
				fields.append("%s=%r" %(field.name, getattr(self, name)))
		return "%s(%s)" %(self.__class__.__name__, ", ".join(fields))
	
	def reset(self):
		for name, field in inspect.get_annotations(self.__class__).items():
			setattr(self, name, None)
	
	def parse(self, data):
		annotations = inspect.get_annotations(self.__class__)

		for name, field in annotations.items():
			setattr(self, name, self.parse_value(field, data.get(field.name)))
		
		names = [field.name for field in annotations.values()]
		for field in data:
			if field not in names:
				raise ValidationError("Received unexpected field '%s'" %field)
	
	def parse_value(self, field, value, is_nested=False):
		if value is None:
			if field.required:
				raise ValidationData("Received null value on required field '%s'" %field.name)
			return value
		
		if field.repeated and not is_nested:
			if not isinstance(value, list):
				raise ValidationError("Expected a list, got %s" %type(value))
			return [self.parse_value(field, element, True) for element in value]
		
		return field.parse(value)

	def encode(self):
		data = {}
		for name, field in inspect.get_annotations(self.__class__).items():
			value = getattr(self, name)
			if value is not None:
				value = field.encode()
			data[field.name] = value
		return data
	
	@classmethod
	def create(cls, data):
		inst = cls()
		inst.parse(data)
		return inst




import httpx


class ProxyError(Exception):
	def __init__(self, response, errors):
		self.response = response
		self.errors = errors
	
	def __str__(self):
		return str(self.errors)


class ServiceError(Exception):
	def __init__(self, response, messages):
		self.response = response
		self.messages = messages
	
	def __str__(self):
		return str(self.messages)


class ServiceClient:
	def __init__(self, settings):
		self.settings = settings

		self.session = httpx.AsyncClient()

		self.account_number = None
		self.card_number = None
	
	async def request(self, method, path, *, service_version=None, **kwargs):
		headers = {
			"User-Agent": self.settings.make_user_agent(self.account_number, self.card_number)
		}
		if service_version is not None:
			headers["x-aab-serviceversion"] = "v%i" %service_version

		response = await self.session.request(method, self.settings.host + path, headers=headers, **kwargs)
		if response.is_error:
			if "application/json" in response.headers.get("Content-Type", ""):
				data = response.json()
				if "messages" in data:
					raise ServiceError(response, data["messages"])
				if "errors" in data:
					raise ProxyError(response, data["errors"])
			response.raise_for_status()
		
		if "application/json" in response.headers.get("Content-Type", ""):
			return response.json()
	
	async def get(self, path, *, service_version=None, params=None):
		return await self.request("GET", path, service_version=service_version, params=params)
	
	async def put(self, path, *, service_version=None, data=None):
		return await self.request("PUT", path, service_version=service_version, json=data)
	
	async def delete(self, path):
		return await self.request("DELETE", path)

	def set_profile(self, account_number, card_number):
		self.account_number = account_number
		self.card_number = card_number


class AccountsService:
	def __init__(self, client):
		self.client = client
	
	async def get_contracts(
		self, product_groups=None, product_building_blocks=None, include_actions=None,
		include_action_names=None, exclude_blocked=None, exclude_status=None,
		bc_number=None, contract_ids=None
	):
		params = {}
		if product_groups is not None: params["productGroups"] = ",".join(product_groups)
		if product_building_blocks is not None: params["productBuildingBlocks"] = ",".join([str(i) for i in product_building_blocks])
		if include_actions is not None: params["includeActions"] = include_actions
		if include_action_names is not None: params["includeActionNames"] = ",".join(include_action_names)
		if exclude_blocked is not None: params["excludeBlocked"] = "true" if exclude_blocked else "false"
		if exclude_status is not None: params["excludeStatus"] = ",".join(exclude_status)
		if bc_number is not None: params["bcNumber"] = bc_number
		if contract_ids is not None: params["contractIds"] = ",".join(contract_ids)
		return await self.client.get("/contracts", params=params, service_version=2)


class AuthorizationService:
	def __init__(self, client):
		self.client = client
	
	async def get_session(self):
		return await self.client.get("/session")

	async def delete_session(self):
		await self.client.delete("/session")

	async def get_login_challenge(self, account_number, card_number, access_tool_usage, bound_device_index_number=None):
		params = {
			"accountNumber": account_number,
			"cardNumber": card_number,
			"accessToolUsage": access_tool_usage
		}
		if access_tool_usage in ["BOUNDDEVICE_USERPIN", "BOUNDDEVICE_TOUCHIDPIN"]:
			params["boundDeviceIndexNumber"] = bound_device_index_number
		
		return await self.client.get("/session/loginchallenge", params=params, service_version=2)
	
	async def send_login_response(
		self, account_number, card_number, challenge_handle, response, access_tool_usage,
		challenge_device_details, app_id, bound_device_index_number=None, is_jailbroken=None,
		is_bound=None, imei=None, telephone_no=None
	):
		data = {
			"accountNumber": account_number,
			"cardNumber": card_number,
			"challengeHandle": challenge_handle,
			"response": response,
			"accessToolUsage": access_tool_usage,
			"challengeDeviceDetails": challenge_device_details,
			"appId": app_id
		}
		if bound_device_index_number is not None: data["boundDeviceIndexNumber"] = bound_device_index_number
		if is_jailbroken is not None: data["isJailbroken"] = is_jailbroken
		if is_bound is not None: data["isBound"] = is_bound
		if imei is not None: data["imei"] = imei
		if telephone_no is not None: data["telephoneNo"] = telephone_no
		return await self.client.put("/session/loginresponse", data=data, service_version=4)

	async def get_session_handover_challenge(self, access_tool_usage):
		params = {
			"accessToolUsage": access_tool_usage
		}
		return await self.client.get("/session/sessionhandoverchallenge", params=params)


class DebitCardsService:
	def __init__(self, client):
		self.client = client
	
	async def get_debit_cards(self):
		return await self.client.get("/debitcards")
	
	async def get_debit_card(self, key):
		return await self.client.get("/debitcards/%s" %key)


class MutationsService:
	def __init__(self, client):
		self.client = client
	
	async def get_mutations(
		self, account, page_size=20, include_actions="BASIC", last_mutation_key=None,
		most_recent_mutation_key=None, search_text=None, amount_from=None, amount_to=None,
		cd_indicator_amount_from=None, cd_indicator_amount_to=None,
		book_date_from=None, book_date_to=None
	):
		params = {
			"pageSize": page_size,
			"includeActions": include_actions
		}
		if last_mutation_key is not None: params["lastMutationKey"] = last_mutation_key
		if most_recent_mutation_key is not None: params["mostRecentMutationKey"] = most_recent_mutation_key
		if search_text is not None: params["searchText"] = search_text
		if amount_from is not None: params["amountFrom"] = "%.2f" %amount_from
		if amount_to is not None: params["amountTo"] = "%.2f" %amount_to
		if cd_indicator_amount_from is not None: params["cdIndicatorAmountFrom"] = cd_indicator_amount_from
		if cd_indicator_amount_to is not None: params["cdIndicatorAmountTo"] = cd_indicator_amount_to
		if book_date_from != None: params["bookDateFrom"] = book_date_from.strftime("%Y/%m/%d")
		if book_date_to != None: params["bookDateTo"] = book_date_to.strftime("%Y/%m/%d")
		return await self.client.get("/mutations/%s" %account, params=params, service_version=3)




import setuptools

long_description = \
	"This is an unofficial Python package that implements the ABN AMRO API."

setuptools.setup(
	name = "abnamro",
	version = "2.0.1",
	description = "Unofficial ABN AMRO client",
	long_description = long_description,
	author = "Yannik Marchand",
	author_email = "ymarchand@me.com",
	url = "https://github.com/kinnay/ABN-Amro",
	packages = ["abnamro", "abnamro.managers"],
	install_requires = [
		"httpx ~= 0.25",
		"pycryptodome"
	]
)




from abnamro import constants

import uuid


USER_AGENT_FORMAT = "[%s]/[%s] [%s]/[%s] [%s]/[%s] [%s] [%s] [%s]"


class Settings:
	def __init__(self):
		self.host = "https://www.abnamro.nl"
		self.app_id = constants.APP_ID_IPHONE

		# User agent settings
		self.app_name = "Bankieren"
		self.app_version = "12.44"
		self.brand_name = "Apple"
		self.model_name = "iPhone15,4"
		self.platform_name = "iOS"
		self.release_name = "17.1.1"
		self.installation_id = str(uuid.uuid4()).upper()
	
	def make_user_agent(self, account_number=None, card_number=None):
		if account_number is None: account_number = ""
		if card_number is None: card_number = ""

		return USER_AGENT_FORMAT %(
			self.app_name, self.app_version, self.brand_name, self.model_name,
			self.platform_name, self.release_name, self.installation_id,
			account_number, card_number
		)




from abnamro import challenge, constants, services
from abnamro.schema import *


class Representative(Schema):
	represetative_class: StringField("class")
	reference: StringField("reference")


class Session(Schema):
	connection_type: StringField("connectionType")
	device_type: StringField("deviceType")
	last_logon_date: TimestampField("lastLogonDate")
	representative: ObjectField("representative", Representative)
	represented_customer: StringField("representedCustomer")
	selected_customer: StringField("selectedCustomer")


class AuthorizationManager:
	def __init__(self, client, settings):
		self.client = client
		self.settings = settings

		self.service = services.AuthorizationService(client)
	
	async def login(self, account_number, card_number, password):
		challenge_result = await self.service.get_login_challenge(account_number, card_number, constants.ACCESS_SOFTTOKEN)
		login_challenge = challenge_result["loginChallenge"]

		challenge_response = challenge.solve_login_challenge(bytes.fromhex(login_challenge["challenge"]), login_challenge["userId"], password)
		login_result = await self.service.send_login_response(
			account_number, card_number, login_challenge["challengeHandle"], challenge_response,
			constants.ACCESS_SOFTTOKEN, login_challenge["challengeDeviceDetails"], self.settings.app_id,
			0, False, False, "", ""
		)
		self.client.set_profile(account_number, card_number)
		return Session.create(login_result["session"])
